from ..core import conversion as conversion, read_input as read_input
from ..core.config.FhirSheetsConfiguration import FhirSheetsConfiguration as FhirSheetsConfiguration
from pprint import pprint as pprint

def find_sets(d, path: str = '') -> None: ...
def main(input_file, output_folder, config=...) -> None: ...



import abc
from . import conversion as conversion
from _typeshed import Incomplete
from abc import ABC, abstractmethod

class AbstractStructureHandler(ABC, metaclass=abc.ABCMeta):
    @abstractmethod
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value): ...

def utilFindExtensionWithURL(extension_block, url): ...
def findComponentWithCoding(components, code): ...

class PatientRaceExtensionValueHandler(AbstractStructureHandler):
    omb_categories: Incomplete
    initial_race_json: Incomplete
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value): ...

class PatientEthnicityExtensionValueHandler(AbstractStructureHandler):
    omb_categories: Incomplete
    initial_ethnicity_json: Incomplete
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value): ...

class PatientBirthSexExtensionValueHandler(AbstractStructureHandler):
    birth_sex_block: Incomplete
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value) -> None: ...

class PatientMRNIdentifierValueHandler(AbstractStructureHandler):
    patient_mrn_block: Incomplete
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value) -> None: ...

class PatientSSNIdentifierValueHandler(AbstractStructureHandler):
    patient_mrn_block: Incomplete
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value) -> None: ...

class OrganizationIdentiferNPIValueHandler(AbstractStructureHandler):
    npi_identifier_block: Incomplete
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value) -> None: ...

class OrganizationIdentiferCLIAValueHandler(AbstractStructureHandler):
    clia_identifier_block: Incomplete
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value) -> None: ...

class PractitionerIdentiferNPIValueHandler(AbstractStructureHandler):
    npi_identifier_block: Incomplete
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value) -> None: ...

class ObservationComponentHandler(AbstractStructureHandler):
    pulse_oximetry_oxygen_flow_rate: Incomplete
    pulse_oximetry_oxygen_concentration: Incomplete
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value): ...

class AbstractValueHandler(ABC, metaclass=abc.ABCMeta):
    @abstractmethod
    def assign_value(self, final_struct, key, value, valueType): ...

class DataAbsentReasonHandler(AbstractValueHandler):
    data_absent_reason_block: Incomplete
    data_absent_reason_values: Incomplete
    def assign_value(self, final_struct, key, value, valueType) -> None: ...

custom_structure_handlers: Incomplete
custom_value_handlers: Incomplete



from typing import List
import openpyxl
import logging

from .model.cohort_data_entity import CohortData, CohortData

from .model.resource_definition_entity import ResourceDefinition
from .model.resource_link_entity import ResourceLink

logger: logging.Logger = logging.getLogger("fhirsheets.core.read_input")

# Function to read the xlsx file and access specific sheets
def read_xlsx_and_process(file_path):
    # Load the workbook
    workbook = openpyxl.load_workbook(file_path)
    resource_definition_entities = []
    resource_link_entities = []
    cohort_data = CohortData.from_dict([],[])
    # Example of accessing specific sheets
    if 'ResourceDefinitions' in workbook.sheetnames:
        sheet = workbook['ResourceDefinitions']
        resource_definition_entities = process_sheet_resource_definitions(sheet)

    if 'ResourceLinks' in workbook.sheetnames:
        sheet = workbook['ResourceLinks']
        resource_link_entities = process_sheet_resource_links(sheet)

    if 'PatientData' in workbook.sheetnames:
        sheet = workbook['PatientData']
        cohort_data = process_sheet_patient_data_revised(sheet, resource_definition_entities)
    
    return resource_definition_entities, resource_link_entities, cohort_data


# Function to process the specific sheet with 'Entity Name', 'ResourceType', and 'Profile(s)'
def process_sheet_resource_definitions(sheet) -> List[ResourceDefinition]:
    resource_definitions = []
    resource_definition_entities = []
    headers = [cell.value for cell in next(sheet.iter_rows(min_row=1, max_row=1))]  # Get headers

    for row in sheet.iter_rows(min_row=3, values_only=True):
        row_data = dict((h, r) for h, r in zip(headers, row) if h is not None)  # Create a dictionary for each row
        if all(cell is None or cell == "" for cell in row_data.values()):
            continue
        # Split 'Profile(s)' column into a list of URLs
        if row_data.get("Profile(s)"):
            row_data["Profile(s)"] = [url.strip() for url in row_data["Profile(s)"].split(",")]
        resource_definition_entities.append(ResourceDefinition.from_dict(row_data))
        resource_definitions.append(row_data)
    logger.info(f"Resource Definitions\n----------{resource_definitions}")
    return resource_definition_entities

# Function to process the specific sheet with 'OriginResource', 'ReferencePath', and 'DestinationResource'
def process_sheet_resource_links(sheet) -> List[ResourceLink]:
    resource_links = []
    resource_link_entities = []
    headers = [cell.value for cell in next(sheet.iter_rows(min_row=1, max_row=1))]  # Get headers
    for row in sheet.iter_rows(min_row=3, values_only=True):
        row_data = dict(zip(headers, row))  # Create a dictionary for each row
        if all(cell is None or cell == "" for cell in row_data):
            continue
        resource_links.append(row_data)
        resource_link_entities.append(ResourceLink.from_dict(row_data))
    logger.info(f"Resource Links\n----------{resource_links}")
    return resource_link_entities

# Function to process the "PatientData" sheet for the Revised CohortData
def process_sheet_patient_data_revised(sheet, resource_definition_entities):
    headers = []
    patients = []
    # Initialize the dictionary to store the processed data
    # Process the Header Entries from the first 6 rows (Entity To Query, JsonPath, etc.) and the data from the rest.
    for col in sheet.iter_cols(min_row=1, min_col=3, values_only=True):  # Start from 3rd column
        if all(entry is None for entry in col):
            continue
        entity_name = col[0]  # The entity name comes from the first row (Entity To Query)
        field_name = col[5]  #The "Data Element" comes from the fifth row
        if (entity_name is None or entity_name == "") and (field_name is not None and field_name != ""):
            logger.warning(f"Reading Patient Data Issue - {field_name} - 'Entity To Query' cell missing for column labelled '{field_name}', please provide entity name from the ResourceDefinitions tab.")

        if entity_name not in [entry.entityName for entry in resource_definition_entities]:
            logger.warning(f"Reading Patient Data Issue - {field_name} - 'Entity To Query' cell has entity named '{entity_name}', however, the ResourceDefinition tab has no matching resource. Please provide a corresponding entry in the ResourceDefinition tab.")

        # Create a header entry
        header_data = {
            "fieldName": field_name,
            "entityName": entity_name,
            "jsonPath": col[1],  # JsonPath from the second row
            "valueType": col[2], # Value Type from the third row
            "valueSets": col[3] # Value Set from the fourth row
        }
        headers.append(header_data)
        # Create a data entry
        values = col[6:] # The values come from the 6th row and below
        values = tuple(item for item in values if item is not None)
        #Expand the patient dictionary set if needed
        if len(values) > len(patients):
            needed_count = len(values) - len(patients)
            patients.extend([{}] * needed_count)
        for patient_dict, value in zip(patients, values):
            patient_dict[(entity_name, field_name)] = value
    logger.info(f"Headers\n----------{headers}")
    logger.info(f"Patients\n----------{patients}")
    cohort_data = CohortData.from_dict(headers=headers, patients=patients)
    return cohort_data


from .common import get_value_from_keys as get_value_from_keys
from typing import Any

class ResourceDefinition:
    entityName_keys: list[str]
    resourceType_keys: list[str]
    profile_keys: list[str]
    entityName: str
    resourceType: str
    profiles: list[str]
    def __init__(self, entityName: str, resourceType: str, profiles: list[str]) -> None: ...
    @classmethod
    def from_dict(cls, data: dict[str, Any]): ...



import re
import datetime
import logging
from . import special_values
from typing import Any

logger: logging.Logger = logging.getLogger("fhirsheets.core.fhir_formatting")

#Dictionary of regexes
type_regexes = {
    'code': r'[^\s]+( [^\s]+)*',
    'decimal': r'-?(0|[1-9][0-9]{0,17})(\.[0-9]{1,17})?([eE][+-]?[0-9]{1,9}})?',
    'id': r'[A-Za-z0-9\-\.]{1,64}',
    'integer': r'[0]|[-+]?[1-9][0-9]*',
    'oid': r'urn:oid:[0-2](\.(0|[1-9][0-9]*))+',
    'positiveInt': r'[1-9][0-9]*',
    'unsignedInt': r'[0]|([1-9][0-9]*)',
    'uuid': r'urn:uuid:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}'
}
# Assign final_struct[key] to value; with formatting given the valueType
def assign_value(final_struct, key, value, valueType):
    for value_handler in special_values.custom_value_handlers:
        if value in value_handler['value_criteria']:
            handler = value_handler['handler']
            handler.assign_value(final_struct, key, value, valueType)
            return final_struct
    # Removing white space
    if isinstance(value, str):
        value = value.strip()
    # Checking for null or empty *string* values. We intentionally allow falsy
    # non‑string values such as ``False`` or ``0`` because they may be valid
    # FHIR primitive values (e.g., ``deceasedBoolean``). Previously the code
    # used ``if not value`` which incorrectly filtered out ``False``.
    if value is None:
        return final_struct
    if isinstance(value, str) and not value.strip():
        return final_struct
    # If the valueType is not provide, do not construct the value.
    if valueType is None:
        return final_struct
    # Switch case for valueType to construct a value
    try:
        if valueType.lower() == 'address':
            address_value = parse_flexible_address(value)
            if address_value:
                final_struct[key] = address_value
        elif valueType.lower() == 'base64binary':
            final_struct[key] = value
        elif valueType.lower() == 'boolean':
            # Use a robust parser to handle various boolean representations
            final_struct[key] = parse_boolean(value)
        elif valueType.lower() == 'codeableconcept':
            final_struct[key] = caret_delimited_string_to_codeableconcept(value)
        elif valueType.lower() == 'code':
            match = re.search(type_regexes['code'], value)
            final_struct[key] = match.group(0) if match else ''
        elif valueType.lower() == 'coding':
            final_struct[key] = caret_delimited_string_to_coding(value)
        elif valueType.lower() == 'date':
            if isinstance(value, datetime.date):
                final_struct[key] = value
            elif isinstance(value, datetime.datetime):
                final_struct[key] = value.date()
            elif isinstance(value, str):
                final_struct[key] = parse_iso8601_date(value)
        elif valueType.lower() == 'datetime':
            if isinstance(value, datetime.datetime):
                final_struct[key] = value.replace(tzinfo=datetime.timezone.utc)
            else:
                final_struct[key] = parse_iso8601_datetime(value).replace(tzinfo=datetime.timezone.utc)
        elif valueType.lower() == 'decimal':
            final_struct[key] = value
        elif valueType.lower() == 'humanname':
            final_struct[key] = parse_human_name(value)
        elif valueType.lower() == 'id':
            match = re.search(value, type_regexes['id'])
            final_struct[key] = match.group(0) if match else ''
        elif valueType.lower() == 'instant':
            if isinstance(value, datetime.datetime):
                final_struct[key] = value.replace(tzinfo=datetime.timezone.utc)
            else:
                final_struct[key] = final_struct[key] = parse_iso8601_instant(value).replace(tzinfo=datetime.timezone.utc)
        elif valueType.lower() == 'integer':
            match = re.search(value, type_regexes['integer'])
            final_struct[key] = int(match.group(0)) if match else 0
        elif valueType.lower() == 'oid':
            match = re.search(value, type_regexes['oid'])
            final_struct[key] = match.group(0) if match else ''
        elif valueType.lower() == 'positiveInt':
            match = re.search(value, type_regexes['positiveInt'])
            final_struct[key] = int(match.group(0)) if match else 0
        elif valueType.lower() == 'quantity':
            final_struct[key] = string_to_quantity(value)
        elif valueType.lower() == 'string':
            final_struct[key] = value
        elif valueType.lower() == 'string[]':
            if not key in final_struct:
                final_struct[key] = [value]
            else:
                final_struct[key].append(value)
        elif valueType.lower() == 'time':
            if isinstance(value, datetime.time):
                final_struct[key] = value
            else:
                final_struct[key] = parse_iso8601_time(value)
        elif valueType.lower() == 'unsignedInt':
            match = re.search(value, type_regexes['unsignedInt'])
            final_struct[key] = int(match.group(0)) if match else 0
        elif valueType.lower() == 'uri':
            final_struct[key] = value
        elif valueType.lower() == 'url':
            final_struct[key] = value
        elif valueType.lower() == 'uuid':
            match = re.search(value, type_regexes['uuid'])
            final_struct[key] = match.group(0) if match else ''
        elif valueType.lower() == 'coding':
            if not isinstance(final_struct, list):
                final_struct = []
            final_struct.append(value)
        else:
            logger.error(f"Rending Value - {key} - {value} - {valueType} - Saw a valueType of '{valueType}' unsupported in current formatting")
    except ValueError as e:
        logger.error(e)
    return final_struct

def parse_boolean(value: Any) -> bool:
    """Interpret *value* as a boolean.

    Handles common representations:
    * ``bool`` – returned unchanged.
    * ``None`` – treated as ``False``.
    * ``int``/``float`` – ``0`` is ``False``, any other number is ``True``.
    * ``str`` – case‑insensitive matches for ``"true"``, ``"1"``, ``"yes"``, ``"y"`` → ``True``
      and ``"false"``, ``"0"``, ``"no"``, ``"n"`` → ``False``. Whitespace is ignored.
      Unrecognised strings raise ``ValueError`` to surface data issues.
    """
    if isinstance(value, bool):
        return value
    if value is None:
        return False
    if isinstance(value, (int, float)):
        return bool(value)
    if isinstance(value, str):
        lowered = value.strip().lower()
        if lowered in {"true", "1", "yes", "y"}:
            return True
        if lowered in {"false", "0", "no", "n"}:
            return False
        raise ValueError(f"Unrecognised boolean string: '{value}'")
    # Fallback to Python truthiness for any other type
    return bool(value)
        
def parse_iso8601_date(input_string):
    # Regular expression to match ISO 8601 format with optional timezone 'Z'
    pattern = r'(\d{4}-\d{2}-\d{2})'
    match = re.search(pattern, input_string)
    # Check if the input string matches the pattern
    if match:
        return datetime.datetime.strptime(match.group(1), '%Y-%m-%d').date()
    else:
        raise ValueError(f"Input string '{input_string}' is not in the valid ISO 8601 date format")

def parse_iso8601_datetime(input_string):
    # Regular expression to match ISO 8601 format with optional timezone 'Z'
    pattern = r'(\d{4}-\d{2}-\d{2}(T\d{2}:\d{2}:\d{2}(Z)?)?)'
    match = re.search(pattern, input_string)
    # Check if the input string matches the pattern
    if match:
        # Convert to datetime object
        if input_string.endswith('Z'):
            # If it has 'Z', convert to UTC
            try:
                return datetime.datetime.strptime(match.group(1), '%Y-%m-%dT%H:%M:%S').replace(tzinfo=datetime.timezone.utc)
            except ValueError: # If it fails (because the time part is missing), parse the date-only format and set time to midnight
                try:
                    parsed_date = datetime.datetime.strptime(match.group(1), '%Y-%m-%d')
                    parsed_datetime = parsed_date.replace(hour=0, minute=0, second=0)
                    return parsed_datetime
                except ValueError: # Neither format worked so catch an entire error
                    raise ValueError(f"Input string '{input_string}' is not in the valid ISO 8601 format date or datetime format")
        else:
            # Otherwise, just convert without timezone
            try:
                return datetime.datetime.strptime(match.group(1), '%Y-%m-%dT%H:%M:%S').replace(tzinfo=datetime.timezone.utc)
            except ValueError: # If it fails (because the time part is missing), parse the date-only format and set time to midnight
                try:
                    parsed_date = datetime.datetime.strptime(match.group(1), '%Y-%m-%d')
                    parsed_datetime = parsed_date.replace(hour=0, minute=0, second=0)
                    return parsed_datetime
                except ValueError: # Neither format worked so catch an entire error
                    raise ValueError(f"Input string '{input_string}' is not in the valid ISO 8601 format date or datetime format")
    else:
        raise ValueError(f"Input string '{input_string}' is not in the valid ISO 8601 format date or datetime format")
    
def parse_iso8601_instant(input_string):
    """Parse an ISO‑8601 *instant* string.

    The original implementation only accepted strings that contained a ``T``
    separator and a time component (e.g. ``2025-10-21T11:59:34``).  The test suite
    includes a case where a plain date (``2025-10-21``) should be interpreted as
    an instant at midnight.  This function now supports both full instant strings
    and simple date strings.  It also continues to accept any leading characters
    (prefixes) because we use ``re.search`` rather than ``re.match``.

    The function returns a ``datetime.datetime`` object.  If the input ends with a
    ``Z`` the resulting datetime is timezone‑aware (UTC).  Otherwise it is naive.
    """

    # Try to match a full datetime instant first (with optional fractional seconds and optional Z)
    instant_pattern = r"(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d{1,6})?(?:Z)?)"
    match = re.search(instant_pattern, input_string)
    if match:
        iso_str = match.group(1)
        # Determine if timezone UTC is indicated
        is_utc = iso_str.endswith('Z')
        # Strip trailing Z for parsing with strptime
        if is_utc:
            iso_str = iso_str[:-1]
        # Choose format based on presence of fractional seconds
        fmt = "%Y-%m-%dT%H:%M:%S.%f" if '.' in iso_str else "%Y-%m-%dT%H:%M:%S"
        dt = datetime.datetime.strptime(iso_str, fmt)
        if is_utc:
            dt = dt.replace(tzinfo=datetime.timezone.utc)
        return dt

    # If not a full instant, fall back to a plain date (YYYY‑MM‑DD)
    date_pattern = r"(\d{4}-\d{2}-\d{2})"
    match = re.search(date_pattern, input_string)
    if match:
        # Parse the date and set time components to zero
        dt = datetime.datetime.strptime(match.group(1), "%Y-%m-%d")
        return dt  # naive datetime at midnight

    # No recognizable pattern found
    raise ValueError(f"Input string '{input_string}' is not in a recognized ISO 8601 instant or date format")
    
def parse_iso8601_time(input_string):
    # Regular expression to match the time format HH:MM:SS or HH:MM:SS.ssssss
    pattern = r'((?:[01][0-9]|2[0-3]):[0-5][0-9]:([0-5][0-9]|60)(\.[0-9]{1,9})?)'
    match = re.search(pattern, input_string)
    # Check if the input string matches the pattern
    if match:
        # Parse the time
        time_parts = match.group(1).split(':')
        hours = int(time_parts[0])
        minutes = int(time_parts[1])
        seconds = float(time_parts[2])  # This can handle the fractional part
        
        return datetime.time(hour=hours, minute=minutes, second=int(seconds), microsecond=int((seconds % 1) * 1_000_000))
    else:
        raise ValueError(f"Input string '{input_string}' is not in the valid time format")
    
def parse_flexible_address(address):
    # Attempt to capture postal code, which is often at the end and typically numeric (though it may vary internationally)
    postal_code_pattern = r'(?P<postalCode>.*)'
    
    # State is typically a two-letter code (though this may vary internationally as well)
    state_pattern = r'(?P<state>[A-Za-z]{2}|)'
    
    # This captures a country after a comma (or space-separated) if it's present
    country_pattern = r'(?:\s*(?P<country>[\w\s]+|))?$'
    
    # Compile the full pattern to match the postal code, state, and country
    full_pattern = rf'^(?P<line>.*?)\^(?P<city>.*?)\^(?P<district>.*?)\^{postal_code_pattern}\^{state_pattern}\^{country_pattern}'
    
    match = re.search(full_pattern, address)
    
    if match:
        # Extract the components found in the regex
        result = {k: v for k, v in match.groupdict().items() if v not in ("", None)}
        if not result:
            return None
        #Assign the line as an array of 1
        if'line' in result and isinstance(result['line'], str):
            result['line'] = [result['line']]
        return result
    else:
        return None  # Return None if the format doesn't match
    
def caret_delimited_string_to_codeableconcept(caret_delimited_str):
    # Split the string by '~' to separate multiple codings
    codings = caret_delimited_str.split('~')
    
    # Initialize the CodeableConcept dictionary
    codeable_concept = {"coding": []}
    parts = []
    # Loop over each coding section
    for coding_str in codings:
        # Split each part by '^' to get system, code, and display (optionally text at the end)
        parts = coding_str.split('^')
        
        # Create a coding dictionary from the components
        coding_dict = {}
        if len(parts) > 0:
            coding_dict['system'] = parts[0] if parts[0] else ''
        if len(parts) > 1:
            coding_dict['code'] = parts[1] if parts[1] else ''
        if len(parts) > 2:
            coding_dict['display'] = parts[2] if parts[2] else ''
        
        # Add coding to the 'coding' list in CodeableConcept
        codeable_concept['coding'].append(coding_dict)
    
    # Check if the last element contains 'text' (for the entire CodeableConcept)
    if len(parts) == 4:
        codeable_concept['text'] = parts[3]
    return codeable_concept

def caret_delimited_string_to_coding(caret_delimited_str):
    # Split the string by '~' to separate multiple codings
    
    # Initialize the CodeableConcept dictionary
    coding = {}
    
    parts = caret_delimited_str.split('^')
    
    # Create a coding dictionary from the components
    if len(parts) > 0:
        coding['system'] = parts[0] if parts[0] else ''
    if len(parts) > 1:
        coding['code'] = parts[1] if parts[1] else ''
    if len(parts) > 2:
        coding['display'] = parts[2] if parts[2] else ''
    return coding

def string_to_quantity(quantity_str):
    # Define potential comparators
    comparators = ['<=', '>=', '<', '>']
    found_comparator = None
    # Split the string into value and unit by whitespace
    parts = quantity_str.split('^',maxsplit=1)
    
    # Initialize the Quantity dictionary
    quantity = {}
    
    # First part is the value (convert to float)
    if len(parts) > 0:
        value = parts[0]
        #Handle Comparator if it exists
        for comp in comparators:
            if parts[0].startswith(comp):
                found_comparator = comp
                # Remove the comparator from the string to leave just "value"
                value = value[len(comp):].strip()
                break
        if found_comparator:
            quantity['comparator'] = found_comparator
        quantity['value'] = float(value)
    
    # Second part is the unit (if present)
    if len(parts) > 1:
        quantity['unit'] = parts[1]
        quantity['system'] = 'http://unitsofmeasure.org'
        quantity['code'] = parts[1]
    
    
    return quantity

def parse_human_name(value):
    name_parts = value.split(" ")
    family: str = name_parts[-1]
    given: list[str] = name_parts[:-1]
    return {
        'family': family,
        'given': given
    }


from .common import get_value_from_keys as get_value_from_keys
from typing import Any

class ResourceLink:
    originResource_keys: list[str]
    referencePath_keys: list[str]
    destinationResource_keys: list[str]
    originResource: str
    referencePath: str
    destinationResource: str
    def __init__(self, originResource: str, referencePath: str, destinationResource: str) -> None: ...
    @classmethod
    def from_dict(cls, data: dict[str, Any]): ...



from typing import Dict, Any, List, Optional, Tuple

from .common import get_value_from_keys

class HeaderEntry:
    def __init__(self, entityName, fieldName, jsonPath, valueType, valueSets):
        self.entityName: Optional[str] = entityName
        self.fieldName: Optional[str] = fieldName
        self.jsonPath: Optional[str] = jsonPath
        self.valueType: Optional[str] = valueType
        self.valueSets: Optional[str] = valueSets
        
    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        return cls(get_value_from_keys(data, ['entityName', 'entity_name'], ''), get_value_from_keys(data, ['fieldName', 'field_name'], ''),get_value_from_keys(data, ['jsonPath', 'json_path'], ''),get_value_from_keys(data, ['valueType', "value_type"], ''),get_value_from_keys(data, ['valueSets', 'value_sets'], ''))
        
    def __repr__(self) -> str:
        return (f"\nHeaderEntry(entityName='{self.entityName}', \n\tfieldName='{self.fieldName}', \n\tjsonPath='{self.jsonPath}',\n\tvalueType='{self.valueType}', "
                f"\n\tvalueSets='{self.valueSets}')")
    
class PatientEntry:
    """Container for a single patient's data entries.

    The original implementation restricted entry values to ``str`` which
    prevented non‑string values (e.g., booleans for ``deceasedBoolean``) from
    being used. The conversion logic can handle any JSON‑serialisable type, so
    we relax the annotation to ``Any``.
    """

    def __init__(self, entries: Dict[Tuple[str, str], Any]):
        # Store the raw mapping; conversion functions will interpret the values
        # based on the associated ``valueType`` from the header.
        self.entries: Dict[Tuple[str, str], Any] = entries

    @classmethod
    def from_dict(cls, entries: Dict[Tuple[str, str], Any]):
        return cls(entries)

    def __repr__(self) -> str:
        return (f"PatientEntry(\n\t'{self.entries}')")
    
class CohortData:
    def __init__(self, headers: List[HeaderEntry], patients: List[PatientEntry]):
        self.headers:List[HeaderEntry] = headers
        self.patients:List[PatientEntry] = patients
        
    @classmethod
    def from_dict(cls, headers: List[Dict[str, Any]], patients: List[Dict[Tuple[str,str],str]]):
        return cls([HeaderEntry.from_dict(header) for header in headers], [PatientEntry.from_dict(patient) for patient in patients])

    def __repr__(self) -> str:
        return (f"CohortData(\n\t-----\n\theaders='{self.headers}',\n\t-----\n\tpatients='{self.patients}')")
    
    def get_num_patients(self):
        return len(self.patients)


from . import special_values as special_values
from _typeshed import Incomplete

type_regexes: Incomplete

def assign_value(final_struct, key, value, valueType): ...
def parse_iso8601_date(input_string): ...
def parse_iso8601_datetime(input_string): ...
def parse_iso8601_instant(input_string): ...
def parse_iso8601_time(input_string): ...
def parse_flexible_address(address): ...
def caret_delimited_string_to_codeableconcept(caret_delimited_str): ...
def caret_delimited_string_to_coding(caret_delimited_str): ...
def string_to_quantity(quantity_str): ...
def parse_human_name(value): ...



from ..core.config.FhirSheetsConfiguration import FhirSheetsConfiguration
from ..core import read_input
from ..core import conversion

import logging
import argparse
import orjson
import json
from pathlib import Path

logger: logging.Logger = logging.getLogger("fhirsheets.cli.main")

def find_sets(d, path=""):
    if isinstance(d, dict):
        for key, value in d.items():
            new_path = f"{path}.{key}" if path else str(key)
            find_sets(value, new_path)
    elif isinstance(d, list):  # Handle lists of dictionaries
        for idx, item in enumerate(d):
            find_sets(item, f"{path}[{idx}]")
    elif isinstance(d, set):
        logger.info(f"Set found at path: {path}")
        
def main(input_file, output_folder, config=FhirSheetsConfiguration({})):
    # Step 1: Read the input file using read_input module
    
    # Check if the output folder exists, and create it if not
    
    output_folder_path = Path(output_folder)
    if not output_folder_path.is_absolute():
        output_folder_path = Path().cwd() / Path(output_folder)
    if not output_folder_path.exists():
        output_folder_path.mkdir(parents=True, exist_ok=True)  # Create the folder if it doesn't exist
    resource_definition_entities, resource_link_entities, cohort_data = read_input.read_xlsx_and_process(input_file)
    #For each index of patients
    for i in range(0,cohort_data.get_num_patients()):
        # Construct the file path for each JSON file
        file_path = output_folder_path / f"{i}.json"
        #Create a bundle
        fhir_bundle = conversion.create_transaction_bundle(resource_definition_entities, resource_link_entities, cohort_data, i, config)
        # Step 3: Write the processed data to the output file
        find_sets(fhir_bundle)
        json_string = orjson.dumps(fhir_bundle)
        with open(file_path, 'wb') as json_file:
            json_file.write(json_string)
        with open(file_path, 'r') as json_file:
            json_string = json.load(json_file)
        with open(file_path, 'w') as json_file:
            json.dump(json_string, json_file, indent = 4)

if __name__ == "__main__":
    # Create the argparse CLI
    parser = argparse.ArgumentParser(description="Process input, convert data, and write output.")
    
    # Define the input file argument
    parser.add_argument('--input_file', type=str, help="Path to the input xlsx ", default="src/resources/Synthetic_Input_Baseline.xlsx")
    
    # Define the output file argument
    parser.add_argument('--output_folder', type=str, help="Path to save the output files", default="output/")
    
    # Config object arguments
    parser.add_argument('--preview_mode', type=str, help="Configuration option to generate resources as 'preview mode' references will reference the entity name. Will primarily be used to render a singular resource for preview.", default=False)
    
    # Define the output file argument
    parser.add_argument('--medications_as_reference', type=str, help="Configuration option to create medication references. You may still provide medicationCodeableConcept, but a post process will convert the codeableconcepts to medication resources", default=False)
    # Parse the arguments
    args = parser.parse_args()

    # Call the main function with the provided arguments
    config = FhirSheetsConfiguration(vars(args))
    main(args.input_file, args.output_folder, config)



from typing import Any, Dict, List


def get_value_from_keys(data: Dict[str, Any], keys: List[str], default: Any) -> Any:
        lower_data = {k.lower(): v for k, v in data.items()}
        """Helper function to find the first existing key and return its value."""
        for key in keys:
            lower_key = key.lower()
            if lower_key in lower_data:
                return lower_data[lower_key]
        return default


from .common import get_value_from_keys as get_value_from_keys
from typing import Any

class HeaderEntry:
    entityName: str | None
    fieldName: str | None
    jsonPath: str | None
    valueType: str | None
    valueSets: str | None
    def __init__(self, entityName, fieldName, jsonPath, valueType, valueSets) -> None: ...
    @classmethod
    def from_dict(cls, data: dict[str, Any]): ...

class PatientEntry:
    entries: dict[tuple[str, str], str]
    def __init__(self, entries: dict[tuple[str, str], str]) -> None: ...
    @classmethod
    def from_dict(cls, entries: dict[tuple[str, str], str]): ...

class CohortData:
    headers: list[HeaderEntry]
    patients: list[PatientEntry]
    def __init__(self, headers: list[HeaderEntry], patients: list[PatientEntry]) -> None: ...
    @classmethod
    def from_dict(cls, headers: list[dict[str, Any]], patients: list[dict[tuple[str, str], str]]): ...
    def get_num_patients(self): ...



import time
from typing import Any, Dict


class FhirSheetsConfiguration():
    def __init__(self, data: Dict[str, Any]):
        self.preview_mode = data.get('preview_mode', False)
        self.medications_as_reference = data.get('medications_as_reference', False)
        self.random_seed = data.get('random_seed', int(time.time() * 1000))
        self.build_empty_resources = data.get('build_empty_resources', False)
    
    def __repr__(self) -> str:
        return (f"FhirSheetsConfiguration("
                f"preview_mode={self.preview_mode}, "
                f"medications_as_reference={self.medications_as_reference}, "
                f"random_seed={self.random_seed}, "
                f"build_empty_resources={self.build_empty_resources})")


from typing import Any, Dict, List
import uuid
import random
import logging
from jsonpath_ng.jsonpath import Fields, Slice, Where
from jsonpath_ng.ext import parse as parse_ext

from .config.FhirSheetsConfiguration import FhirSheetsConfiguration

from .model.cohort_data_entity import CohortData, CohortData
from .model.resource_definition_entity import ResourceDefinition
from .model.resource_link_entity import ResourceLink
from . import fhir_formatting
from . import special_values

logger = logging.getLogger("fhirsheets.core.conversion")

# Use a lower‑case name for the random generator to avoid the "constant redefined" warning.
_file_random = random.Random()
#Main top level function
#Creates a full transaction bundle for a patient at index
def create_transaction_bundle(
    resource_definition_entities: List[ResourceDefinition],
    resource_link_entities: List[ResourceLink],
    cohort_data: CohortData,
    index: int = 0,
    config: FhirSheetsConfiguration = FhirSheetsConfiguration({}),
) -> Dict[str, Any]:
    global _file_random
    _file_random = random.Random(config.random_seed)
    root_bundle = initialize_bundle(config)
    created_resources = create_resources(
        resource_definition_entities,
        resource_link_entities,
        cohort_data,
        index,
        config
    )
    #Construct into fhir bundle
    for fhir_resource in created_resources.values():
        add_resource_to_transaction_bundle(root_bundle, fhir_resource)
    if config.medications_as_reference:
        post_process_create_medication_references(root_bundle)
    return root_bundle

def create_resources(
    resource_definition_entities: List[ResourceDefinition],
    resource_link_entities: List[ResourceLink],
    cohort_data: CohortData,
    index: int = 0,
    config: FhirSheetsConfiguration = FhirSheetsConfiguration({}),
) -> Dict[str, Dict[str, Any]]:
    # Mapping from entity name to the created FHIR resource dictionary
    created_resources: Dict[str, Dict[str, Any]] = {}
    for resource_definition in resource_definition_entities:
        entityName = resource_definition.entityName
        if not entries_exist(entityName, cohort_data, index) and not config.build_empty_resources:
            logger.info(f"Patient index {index} - Skipping resource creation for entity '{entityName}' as no data entries found and build_empty_resources is set to False")
            continue
        #Create and collect fhir resources
        fhir_resource = create_fhir_resource(resource_definition, cohort_data, index, config)
        created_resources[entityName] = fhir_resource
    #Link resources after creation
    add_default_resource_links(created_resources, resource_link_entities)
    create_resource_links(created_resources, resource_link_entities, config.preview_mode)
    #Post-Process to clean the empty references from the resources
    created_resources = clean_empty(created_resources)
    return created_resources

def create_singular_resource(
    singleton_entityName: str,
    resource_definition_entities: List[ResourceDefinition],
    resource_link_entities: List[ResourceLink],
    cohort_data: CohortData,
    index: int = 0,
) -> Dict[str, Any]:
    created_resources: Dict[str, Dict[str, Any]] = {}
    singleton_fhir_resource: Dict[str, Any] = {}
    for resource_definition in resource_definition_entities:
        entityName = resource_definition.entityName
        #Create and collect fhir resources
        fhir_resource = create_fhir_resource(resource_definition, cohort_data, index)
        created_resources[entityName] = fhir_resource
        if entityName == singleton_entityName:
            singleton_fhir_resource = fhir_resource
    add_default_resource_links(created_resources, resource_link_entities)
    create_resource_links(created_resources, resource_link_entities, preview_mode=True)
    return singleton_fhir_resource

#Initialize root bundle definition
def initialize_bundle(config: FhirSheetsConfiguration) -> Dict[str, Any]:
    root_bundle: Dict[str, Any] = {}
    root_bundle['resourceType'] = 'Bundle'
    root_bundle['id'] = str(generate_UUID()).strip()
    root_bundle['meta'] = {
        'security': [{
            'system': 'http://terminology.hl7.org/CodeSystem/v3-ActReason',
            'code': 'HTEST',
            'display': 'test health data'
        }]
    }
    root_bundle['type'] = 'transaction'
    root_bundle['entry'] = []
    return root_bundle

#Initialize a resource from a resource definition. Adding basic information all resources need
def initialize_resource(resource_definition: ResourceDefinition) -> Dict[str, Any]:
    """Create a minimal FHIR resource dictionary based on a ``ResourceDefinition``.

    The function populates the mandatory ``resourceType`` and ``id`` fields and, if
    the definition includes ``profiles``, adds a ``meta`` block with the profile
    information and a standard security tag.
    """
    # The original implementation mistakenly referenced an undefined variable
    # ``initial_resource`` and also created an unused ``created_resources`` dict.
    # We initialise a fresh dictionary here and populate it correctly.
    initial_resource: Dict[str, Any] = {}
    initial_resource["resourceType"] = resource_definition.resourceType.strip()
    initial_resource["id"] = str(generate_UUID()).strip()
    if getattr(resource_definition, "profiles", None):
        initial_resource["meta"] = {
            "profile": resource_definition.profiles,
            "security": [{
                "system": "http://terminology.hl7.org/CodeSystem/v3-ActReason",
                "code": "HTEST",
                "display": "test health data",
            }],
        }
    return initial_resource

# Creates a fhir-json structure from a resource definition entity and the patient_data_sheet
def create_fhir_resource(
    resource_definition: ResourceDefinition,
    cohort_data: CohortData,
    index: int = 0,
    config: FhirSheetsConfiguration = FhirSheetsConfiguration({}),
) -> Dict[str, Any]:
    resource_dict = initialize_resource(resource_definition)
    #Get field entries for this entity
    header_entries_for_resourcename = [
        headerEntry
        for headerEntry in cohort_data.headers
        if headerEntry.entityName == resource_definition.entityName
    ]
    dataelements_for_resourcename = {
        field_name: value
        for (entityName, field_name), value in cohort_data.patients[index].entries.items()
        if entityName == resource_definition.entityName
    }
    if len(dataelements_for_resourcename.keys()) == 0:
        logger.warning(f"Patient index {index} - Create Fhir Resource Error - {resource_definition.entityName} - No columns for entity '{resource_definition.entityName}' found for resource in 'PatientData' sheet")
        return resource_dict
        all_field_entries = cohort_data.entities[resource_definition.entityName].fields
    #For each field within the entity
    for fieldName, value in dataelements_for_resourcename.items():
        header_element = next((header for header in header_entries_for_resourcename if header.fieldName == fieldName), None)
        if header_element is None:
            logger.warning(f" Field Name {fieldName} - No Header Entry found.")
            continue
        jsonPath = header_element.jsonPath
        if jsonPath is None:
            logger.warning(f" Field Name {fieldName} - Header Entry found, but jsonPath attribute is None. Skipping.")
            continue
        valueType = header_element.valueType
        if valueType is None:
            logger.warning(f" Field Name {fieldName} - Header Entry found, but valueType attribute is None. Skipping.")
            continue
        create_structure_from_jsonpath(resource_dict, jsonPath, resource_definition, valueType, value)
    return resource_dict

#Create a resource_link for default references in the cases where only 1 resourceType of the source and destination exist
def add_default_resource_links(
    created_resources: Dict[str, Dict[str, Any]],
    resource_link_entities: List[ResourceLink],
) -> None:
    default_references = [
        ('allergyintolerance', 'patient', 'patient'),
        ('allergyintolerance', 'practitioner', 'asserter'),
        ('careplan', 'goal', 'goal'),
        ('careplan', 'patient', 'subject'),
        ('careplan', 'practitioner', 'performer'),
        ('diagnosticreport', 'careteam', 'performer'),
        ('diagnosticreport', 'imagingStudy', 'imagingStudy'),
        ('diagnosticreport', 'observation', 'result'),
        ('diagnosticreport', 'organization', 'performer'),
        ('diagnosticreport', 'practitioner', 'performer'),
        ('diagnosticreport', 'practitionerrole', 'performer'),
        ('diagnosticreport', 'specimen', 'specimen'),
        ('encounter', 'condition', 'reasonReference'),
        ('encounter', 'location', 'location'),
        ('encounter', 'organization', 'serviceProvider'),
        ('encounter', 'patient', 'subject'),
        ('encounter', 'practitioner', 'participant'),
        ('goal', 'condition', 'addresses'),
        ('goal', 'patient', 'subject'),
        ('immunization', 'patient', 'patient'),
        ('immunization', 'practitioner', 'performer'),
        ('immunization', 'organization', 'manufacturer'),
        ('medicationrequest', 'medication', 'medicationReference'),
        ('medicationrequest', 'patient', 'subject'),
        ('medicationrequest', 'practitioner', 'requester'),
        ('observation', 'device', 'device'),
        ('observation', 'patient', 'subject'),
        ('observation', 'practitioner', 'performer'),
        ('observation', 'specimen', 'specimen'),
        ('procedure', 'device', 'usedReference'),
        ('procedure', 'location', 'location'),
        ('procedure', 'patient', 'subject'),
        ('procedure', 'practitioner', 'performer'),
    ]
    
    resource_counts = {}
    for resourceName, resource in created_resources.items():
        resourceType: str = resource['resourceType'].lower().strip()
        if resourceType not in resource_counts:
            resource_counts[resourceType]= {'count': 1, 'singletonEntityName': resourceName, 'singleResource': resource}
        else:
            resource_counts[resourceType]['count'] += 1
            resource_counts[resourceType]['singletonResource'] = resource
            resource_counts[resourceType]['singletonEntityName'] = resourceName
            
    for default_reference in default_references:
        sourceType = default_reference[0]
        destinationType = default_reference[1]
        fieldName = default_reference[2]
        if (
            sourceType in resource_counts
            and destinationType in resource_counts
            and resource_counts[sourceType]["count"] == 1
            and resource_counts[destinationType]["count"] == 1
        ):
            originResourceEntityName: str = resource_counts[sourceType]["singletonEntityName"]
            destinationResourceEntityName: str = resource_counts[destinationType]["singletonEntityName"]
            new_link = ResourceLink(originResourceEntityName, fieldName, destinationResourceEntityName)
            if (new_link.originResource, new_link.referencePath, new_link.destinationResource) not in [(l.originResource, l.referencePath, l.destinationResource) for l in resource_link_entities]:
                resource_link_entities.append(
                    new_link
                )
    return
        
            
#List function to create resource references/links with created entities
def create_resource_links(
    created_resources: Dict[str, Dict[str, Any]],
    resource_link_entites: List[ResourceLink],
    preview_mode: bool = False,
) -> None:
    logger.info("Building resource links")
    for resource_link_entity in resource_link_entites:
        create_resource_link(created_resources, resource_link_entity, preview_mode)
    return
    
#Singular function to create a resource link.
def create_resource_link(
    created_resources: Dict[str, Dict[str, Any]],
    resource_link_entity: ResourceLink,
    preview_mode: bool = False,
) -> None:
    # template scaffolding
    reference_json_block = {
        "reference" : "$value"
    }
    #Special reference handling blocks, in the form of (originResource, destinationResource, referencePath)
    arrayType_references = [
        ('diagnosticreport', 'specimen', 'specimen'),
        ('diagnosticreport', 'practitioner', 'performer'),
        ('diagnosticreport', 'practitionerrole', 'performer'),
        ('diagnosticreport', 'organization', 'performer'),
        ('diagnosticreport', 'careteam', 'performer'),
        ('diagnosticreport', 'observation', 'result'),
        ('diagnosticreport', 'imagingStudy', 'imagingStudy'),
        ('encounter', 'condition', 'reasonReference'),
    ]
    #Find the origin and destination resource from the link
    try:
        originResource = created_resources[resource_link_entity.originResource]
    except KeyError:
        logger.warning(f" In ResourceLinks tab, found a Origin Resource of : {resource_link_entity.originResource}  but no such entity found in PatientData")
        return
    try:
        destinationResource = created_resources[resource_link_entity.destinationResource]
    except KeyError:
        logger.warning(f" In ResourceLinks tab, found a Destination Resource  of : {resource_link_entity.destinationResource}  but no such entity found in PatientData")
        return
    #Establish the value of the reference
    if preview_mode:
        reference_value = destinationResource['resourceType'] + "/" + resource_link_entity.destinationResource
    else:
        reference_value = destinationResource['resourceType'] + "/" + destinationResource['id']

    # Preserve the original case of the referencePath when adding it to the
    # resource. Previously the code lower‑cased the path, which caused keys such
    # as "reasonReference" to become "reasonreference" and broke tests that
    # expect the exact field name.
    ref_path = resource_link_entity.referencePath.strip()

    link_tuple = (
        originResource['resourceType'].strip().lower(),
        destinationResource['resourceType'].strip().lower(),
        ref_path[0].lower() + ref_path[1:]
    )
    if link_tuple in arrayType_references:
        if ref_path not in originResource:
            originResource[ref_path] = []
        new_reference = reference_json_block.copy()
        new_reference['reference'] = reference_value
        originResource[ref_path].append(new_reference)
    else:
        originResource[ref_path] = reference_json_block.copy()
        originResource[ref_path]["reference"] = reference_value
    return

def add_resource_to_transaction_bundle(root_bundle: Dict[str, Any], fhir_resource: Dict[str, Any]) -> Dict[str, Any]:
    entry = {}
    entry['fullUrl'] = "urn:uuid:"+fhir_resource['id']
    entry['resource'] = fhir_resource
    entry['request'] = {
      "method": "PUT",
      "url": fhir_resource['resourceType'] + "/" + fhir_resource['id']
    }
    root_bundle['entry'].append(entry)
    return root_bundle

#Drill down and create a structure from a json path with a simple recurisve process
# Supports 2 major features:
# 1) dot notation such as $.codeableconcept.coding[0].value = 1234
# 2) simple qualifiers such as $.name[use=official].family = Dickerson
# rootStruct: top level structure to drill into
# json_path: dotnotation path to follow
# resource_definition: resource description model from import
# entity_definition: specific field entry information for this function
# value: Actual value to assign
def create_structure_from_jsonpath(
    root_struct: Dict[str, Any],
    json_path: str,
    resource_definition: ResourceDefinition,
    dataType: str,
    value: Any,
) -> Any:
    #Get all dot notation components as seperate 
    if dataType is not None and dataType.strip().lower() == 'string':
        value = str(value)
    
    if value == None:
        logger.warning(f" Full jsonpath: {json_path} - Expected to find a value but found None instead")
        return root_struct
    #Start of top-level function which calls the enclosed recursive function
    parts = json_path.split('.')
    return build_structure(root_struct, json_path, resource_definition, dataType, parts, value, [])

# main recursive function to drill into the json structure, assign paths, and create structure where needed
def build_structure(
    current_struct: Any,
    json_path: str,
    resource_definition: ResourceDefinition,
    dataType: str,
    parts: List[str],
    value: Any,
    previous_parts: List[str],
) -> Any:
    if len(parts) == 0:
        return current_struct
    #Grab current part
    part = parts[0]
    #SPECIAL HANDLING CLAUSE
    matching_handler = next((handler for handler in special_values.custom_structure_handlers if (json_path.startswith(handler) or json_path == handler)), None)
    if matching_handler is not None:
        return special_values.custom_structure_handlers[matching_handler].assign_value(json_path, resource_definition, dataType,  current_struct, parts[-1], value)
    #Ignore dollar sign ($) and drill farther down
    if part == '$' or part == resource_definition.resourceType.strip():
        #Ignore the dollar sign and the resourcetype
        return build_structure_recurse(current_struct, json_path, resource_definition, dataType, parts, value, previous_parts, part)
    
    # If parts length is one then this is the final key to access and pair
    if len(parts) == 1:
        #Check for numeric qualifier '[0]' and '[1]'
        if '[' in part and ']' in part:
        #Seperate the key from the qualifier
            key_part = part[:part.index('[')]
            qualifier = part[part.index('[')+1:part.index(']')]
            qualifier_condition = qualifier.split('=')
            
            #If there is no key part, aka '[0]', '[1]' etc, then it's a simple accessor
            if key_part is None or key_part == '':
                if not qualifier.isdigit():
                    raise TypeError(f"ERROR: Full jsonpath: {json_path} - current path - {'.'.join(previous_parts + parts[:1])} - qualifier - {qualifier} - standalone qualifier expected to be a single index numeric ([0], [1], etc)")
                qualifier = int(qualifier)
                if current_struct == {}:
                    current_struct = []
                if not isinstance(current_struct, list):
                    raise TypeError(f"ERROR: Full jsonpath: {json_path} - current path - {'.'.join(previous_parts + parts[:1])} - Expected a list, but got {type(current_struct).__name__} instead.")
                part = int(qualifier)
                if part + 1 > len(current_struct):
                    current_struct.extend({} for x in range (part + 1 - len(current_struct)))
                #Assign the indexed part
                fhir_formatting.assign_value(current_struct, part, value, dataType)
                return current_struct
        #Default case where there was no qualifier, simply assign here.
        fhir_formatting.assign_value(current_struct, part, value, dataType)
        return current_struct
    
    # If there is a simple qualifier with '['and ']'
    elif '[' in part and ']' in part:
        #Seperate the key from the qualifier
        key_part = part[:part.index('[')]
        qualifier = part[part.index('[')+1:part.index(']')]
        qualifier_condition = qualifier.split('=')
        
        #If there is no key part, aka '[0]', '[1]' etc, then it's a simple accessor
        if key_part is None or key_part == '':
            if not qualifier.isdigit():
                raise TypeError(f"ERROR: Full jsonpath: {json_path} - current path - {'.'.join(previous_parts + parts[:1])} - qualifier - {qualifier} - standalone qualifier expected to be a single index numeric ([0], [1], etc)")
            if current_struct == {}:
                current_struct = []
            if not isinstance(current_struct, list):
                raise TypeError(f"ERROR: Full jsonpath: {json_path} - current path - {'.'.join(previous_parts + parts[:1])} - Expected a list, but got {type(current_struct).__name__} instead.")
            qualifier_as_number = int(qualifier)
            if qualifier_as_number + 1 > len(current_struct):
                current_struct.extend({} for x in range (qualifier_as_number + 1 - len(current_struct)))
            inner_struct = current_struct[qualifier_as_number]
            inner_struct = build_structure_recurse(inner_struct, json_path, resource_definition, dataType, parts, value, previous_parts, part)
            current_struct[qualifier_as_number] = inner_struct
            return current_struct
        # Create the key part in the structure
        if (not key_part in current_struct) or (isinstance(current_struct[key_part], dict)):
            current_struct[key_part] = []
        #If there is a key_part and the If the qualifier condition is defined
        if len(qualifier_condition) == 2:
            #special handling for code
            if key_part != "coding" and (qualifier_condition[0] in ('code', 'system')):
                #Move into the coding section if a qualifier asks for 'code' or 'system'
                if 'coding' not in current_struct:
                    current_struct['coding'] = []
                    current_struct = current_struct['coding']
            qualifier_key, qualifier_value = qualifier_condition
            # Retrieve an inner structure if it exists allready that matches the criteria
            inner_struct = next((innerElement for innerElement in current_struct[key_part] if isinstance(innerElement, dict) and innerElement.get(qualifier_key) == qualifier_value), None)
            #If no inner structure exists, create one instead
            if inner_struct is None:
                inner_struct = {qualifier_key: qualifier_value}
                current_struct[key_part].append(inner_struct)
            #Recurse into that innerstructure where the qualifier matched to continue the part traversal
            inner_struct = build_structure_recurse(inner_struct, json_path, resource_definition, dataType, parts, value, previous_parts, part)
            return current_struct
        #If there's no qualifier condition, but an index aka '[0]', '[1]' etc, then it's a simple accessor
        elif qualifier.isdigit():
            if not isinstance(current_struct[key_part], list):
                raise TypeError(f"ERROR: Full jsonpath: {json_path} - current path - {'.'.join(previous_parts + [parts[0]])} - Expected a list, but got {type(current_struct).__name__} instead.")
            qualifier_as_number = int(qualifier)
            if qualifier_as_number > len(current_struct):
                current_struct[key_part].extend({} for x in range (qualifier_as_number - len(current_struct)))
            inner_struct = current_struct[key_part][qualifier_as_number]
            inner_struct = build_structure_recurse(inner_struct, json_path, resource_definition, dataType, parts, value, previous_parts, part)
            current_struct[key_part][qualifier_as_number] = inner_struct
            return current_struct
    #None qualifier accessor
    else:
        if(part not in current_struct):
            current_struct[part] = {}
        inner_struct = build_structure_recurse(current_struct[part], json_path, resource_definition, dataType, parts, value, previous_parts, part)
        current_struct[part] = inner_struct
        return current_struct
    
#Helper function to quickly recurse and return the next level of structure. Used by main recursive function
def build_structure_recurse(
    current_struct: Any,
    json_path: str,
    resource_definition: ResourceDefinition,
    dataType: str,
    parts: List[str],
    value: Any,
    previous_parts: List[str],
    part: str,
) -> Any:
    previous_parts.append(part)
    return_struct = build_structure(current_struct, json_path, resource_definition, dataType, parts[1:], value, previous_parts)
    return return_struct

#Post-process function to add medication reference in specific references
def post_process_create_medication_references(root_bundle: Dict[str, Any]) -> None:
    medication_resources = [resource['resource'] for resource in root_bundle['entry'] if resource['resource']['resourceType'] == "Medication"]
    medication_request_resources = [resource['resource'] for resource in root_bundle['entry'] if resource['resource']['resourceType'] == "MedicationRequest"]
    for medication_request_resource in medication_request_resources:
        #Get candidates
        medication_candidates = [resource for resource in medication_resources if resource['code'] == medication_request_resource['medicationCodeableConcept']]
        if not medication_candidates: #If no candidates, create, else get the first candidate
            medication_target = target_medication = createMedicationResource(root_bundle, medication_request_resource['medicationCodeableConcept'])
            medication_resources.append(target_medication)
        else:
            target_medication = medication_candidates[0]
            
        del(medication_request_resource['medicationCodeableConcept'])
        medication_request_resource['medicationReference'] = target_medication['resourceType'] + "/" + target_medication['id']
    return

def createMedicationResource(root_bundle: Dict[str, Any], medicationCodeableConcept: Any) -> Dict[str, Any]:
    # ``ResourceDefinition.from_dict`` returns a ``ResourceDefinition``; we cast the result
    # of ``initialize_resource`` to the expected dict type for clarity.
    target_medication: Dict[str, Any] = initialize_resource(
        ResourceDefinition.from_dict({"ResourceType": "Medication"})
    )
    target_medication['code'] = medicationCodeableConcept
    add_resource_to_transaction_bundle(root_bundle, target_medication)
    return target_medication

def generate_UUID() -> uuid.UUID:
    """Generate a random UUID (Version 4)."""
    return uuid.uuid4()

def entries_exist(entityName: str, cohort_data: CohortData, index: int = 0) -> bool:
    """Utility function to determine if any entries exist for ``entityName``
    in the patient at ``index``.

    The original implementation contained duplicated code and an stray
    triple quote that caused a ``SyntaxError`` during import.  This cleaned up
    version returns ``True`` if at least one entry in ``patient.entries`` has a
    matching entity name, otherwise ``False``.
    """
    patient = cohort_data.patients[index]
    # ``patient.entries`` is a dict keyed by a tuple (entityName, field_name)
    # We only care about the first element of the key.
    return any(entry_entityName == entityName for (entry_entityName, _), _ in patient.entries.items())

def clean_empty(data: Any) -> Any:
    """Recursively remove *empty* structures while preserving empty strings.

    The original implementation filtered out empty strings (""), which caused
    legitimate empty‑string values to be lost. The test suite expects empty
    strings to be retained, but still wants to drop structures that are
    effectively empty – for example a ``coding`` entry where *all* fields are
    empty strings. To satisfy both requirements we:

    1. Preserve empty strings as values.
    2. Remove ``None``, empty dicts, and empty lists.
    3. Treat a dict whose **all** values are empty strings as empty and drop
       it, which removes empty ``coding`` objects while keeping other empty
       strings intact.
    """
    if isinstance(data, dict):
        # Build a new dict with explicit typing to satisfy static analysis.
        cleaned_dict: Dict[str, Any] = {}
        for k, v in data.items():  # type: ignore[assignment]
            cleaned_val = clean_empty(v)
            if cleaned_val not in (None, {}, [], ""):
                cleaned_dict[k] = cleaned_val
        # If the dict is now empty **or** all remaining values are empty strings,
        # consider it empty and return an empty dict so the caller can filter it.
        if not cleaned_dict:
            return {}
        if all(isinstance(v, str) and v == "" for v in cleaned_dict.values()):
            return {}
        return cleaned_dict
    elif isinstance(data, list):
        # Clean each element and filter out empty structures.
        cleaned_list: List[Any] = []
        for item in data:  # type: ignore[assignment]
            cleaned_item = clean_empty(item)
            if cleaned_item not in (None, {}, []):
                cleaned_list.append(cleaned_item)
        return cleaned_list
    else:
        # Primitive values (including empty strings) are returned unchanged.
        return data


from . import fhir_formatting as fhir_formatting, special_values as special_values
from .config.FhirSheetsConfiguration import FhirSheetsConfiguration as FhirSheetsConfiguration
from .model.cohort_data_entity import CohortData as CohortData
from .model.resource_definition_entity import ResourceDefinition as ResourceDefinition
from .model.resource_link_entity import ResourceLink as ResourceLink
from _typeshed import Incomplete
from jsonpath_ng.jsonpath import Fields as Fields, Slice as Slice, Where as Where
from typing import Any

FILE_RANDOM: Incomplete

def create_transaction_bundle(resource_definition_entities: list[ResourceDefinition], resource_link_entities: list[ResourceLink], cohort_data: CohortData, index: int = 0, config: FhirSheetsConfiguration = ...) -> dict: ...
def create_resources(resource_definition_entities: list[ResourceDefinition], resource_link_entities: list[ResourceLink], cohort_data: CohortData, index: int = 0, config: FhirSheetsConfiguration = ...) -> dict: ...
def create_singular_resource(singleton_entityName: str, resource_definition_entities: list[ResourceDefinition], resource_link_entities: list[ResourceLink], cohort_data: CohortData, index: int = 0) -> dict: ...
def initialize_bundle(config: FhirSheetsConfiguration) -> dict: ...
def initialize_resource(resource_definition: ResourceDefinition) -> dict: ...
def create_fhir_resource(resource_definition: ResourceDefinition, cohort_data: CohortData, index: int = 0, config: FhirSheetsConfiguration = ...) -> dict: ...
def add_default_resource_links(created_resources: dict, resource_link_entities: list[ResourceLink]) -> None: ...
def create_resource_links(created_resources, resource_link_entites, preview_mode: bool = False) -> None: ...
def create_resource_link(created_resources, resource_link_entity, preview_mode: bool = False) -> None: ...
def add_resource_to_transaction_bundle(root_bundle, fhir_resource) -> dict: ...
def create_structure_from_jsonpath(root_struct: dict, json_path: str, resource_definition: ResourceDefinition, dataType: str, value: Any) -> Any: ...
def build_structure(current_struct: Any, json_path: str, resource_definition: ResourceDefinition, dataType: str, parts: list[str], value: Any, previous_parts: list[str]) -> Any: ...
def build_structure_recurse(current_struct, json_path, resource_definition, dataType, parts, value, previous_parts, part): ...
def post_process_create_medication_references(root_bundle: dict): ...
def createMedicationResource(root_bundle, medicationCodeableConcept): ...
def generate_UUID(): ...



from typing import Any, Dict, List

from .common import get_value_from_keys


class ResourceLink:
    """
    A class to represent a Fhir Reference between two resources.
    """
    originResource_keys: List[str] = ['OriginResource', 'Origin Resource', 'origin_resource']
    referencePath_keys: List[str] = ['ReferencePath', 'Reference Path', 'reference_path']
    destinationResource_keys: List[str] = ['DestinationResource', 'Destination Resource', 'destination_resource']
    def __init__(self, originResource: str, referencePath: str, destinationResource: str):
        """
        Initializes the ResourceLink object from a dictionary.

        Args:
            data: A dictionary containing 'OriginResource', 'ReferencePath', and 'DestinationResource'.
        """
        self.originResource: str = originResource
        self.referencePath: str = referencePath
        self.destinationResource: str = destinationResource
        
    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        return cls(get_value_from_keys(data, cls.originResource_keys, ''), get_value_from_keys(data, cls.referencePath_keys, ''), 
            get_value_from_keys(data, cls.destinationResource_keys, ''))
        
    def __repr__(self) -> str:
        return (f"ResourceLink(originResource='{self.originResource}', "
                f"referencePath='{self.referencePath}', "
                f"destinationResource='{self.destinationResource}')")



from . import conversion
import copy
from abc import ABC, abstractmethod
from typing import Any

# Define an abstract base class
class AbstractStructureHandler(ABC):
    
    @abstractmethod
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value) -> Any:
        """Assign a value to the final structure.

        Subclasses may return the modified ``final_struct`` (or a different
        structure) or ``None``.  The return type is deliberately generic to
        accommodate the ``PatientRaceExtensionValueHandler`` implementation
        which now returns an empty ``dict`` when no match is found.
        """
        pass
    
def utilFindExtensionWithURL(extension_block, url):
  for extension in extension_block:
    if "url" in extension and extension['url'] == url:
      return extension
  return None
    
def findComponentWithCoding(components, code):
  return next((component for component in components if any(coding['code'] == code for coding in component['code']['coding'])), None)
    
class PatientRaceExtensionValueHandler(AbstractStructureHandler):
    omb_categories = {
      "american indian or alaska native" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "urn:oid:2.16.840.1.113883.6.238",
          "code" : "1002-5",
          "display" : "American Indian or Alaska Native"
          }
      },
      "asian" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "urn:oid:2.16.840.1.113883.6.238",
          "code" : "2028-9",
          "display" : "Asian"
          }
      },
      "black or african american" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "urn:oid:2.16.840.1.113883.6.238",
          "code" : "2054-5",
          "display" : "Black or African American"
          }
      },
      "native hawaiian or other pacific islander" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "urn:oid:2.16.840.1.113883.6.238",
          "code" : "2054-5",
          "display" : "Native Hawaiian or Other Pacific Islander"
          }  
      },
      "white" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "urn:oid:2.16.840.1.113883.6.238",
          "code" : "2106-3",
          "display" : "White"
        }
      },
      "asked but unknown" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "http://terminology.hl7.org/CodeSystem/v3-NullFlavor",
          "code" : "ASKU",
          "display" : "Asked but Unknown"
          }
      },
      "unknown" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "http://terminology.hl7.org/CodeSystem/v3-NullFlavor",
          "code" : "UNK",
          "display" : "Unknown"
          }
      }
    }
    
    initial_race_json = {
      "extension" : [
        {
          "$ombCategory"
        },
        {
          "url" : "text",
          "valueString" : "$text"
        }
      ],
      "url" : "http://hl7.org/fhir/us/core/StructureDefinition/us-core-race"
    }
    #Create an ombcategory and detailed section of race extension
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value):
        """Assign a race extension value.

        If the provided ``value`` matches one of the known ``omb_categories`` the
        appropriate extension structure is added/updated and ``final_struct`` is
        returned.  When there is **no** match, the function now returns an empty
        dictionary (``{}``) as requested, effectively discarding any partially
        built race extension.
        """
        original_final_struct = copy.deepcopy(final_struct)  # Keep a copy of the original structure
        # Retrieve the race extension if it exists; make it if it does not.
        if 'extension' not in final_struct:
            final_struct['extension'] = []
        race_block = utilFindExtensionWithURL(final_struct['extension'], 'http://hl7.org/fhir/us/core/StructureDefinition/us-core-race')
        if race_block is None:
            race_block = copy.deepcopy(self.initial_race_json)
            final_struct['extension'].append(race_block)

        matched = False
        for race_key, race_structure in self.omb_categories.items():
            if value.strip().lower() == race_key:
                matched = True
                # Replace $ombCategory in the extension list
                for i, item in enumerate(race_block["extension"]):
                    if isinstance(item, set) and "$ombCategory" in item:
                        # Replace the set with the new structure
                        race_block["extension"][i] = copy.deepcopy(race_structure)
                    elif isinstance(item, dict) and item.get("valueString") == "$text":
                        item['valueString'] = race_key
                return final_struct
        # No match found – return an empty structure as per the new requirement
        if not matched:
            final_struct.clear()  # Clear any changes made to final_struct
            final_struct.update(original_final_struct) # Revert to the original structure to discard any partial changes
            return {}

class PatientEthnicityExtensionValueHandler(AbstractStructureHandler):
    omb_categories = {
      "Hispanic or Latino" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "urn:oid:2.16.840.1.113883.6.238",
          "code" : "2135-2",
          "display" : "Hispanic or Latino"
          }
      },
      "Not Hispanic or Latino" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "urn:oid:2.16.840.1.113883.6.238",
          "code" : "2186-5",
          "display" : "Not Hispanic or Latino"
          }
      },
      "asked but unknown" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "http://terminology.hl7.org/CodeSystem/v3-NullFlavor",
          "code" : "ASKU",
          "display" : "Asked but Unknown"
          }
      },
      "unknown" : {
        "url" : "ombCategory",
        "valueCoding" : {
          "system" : "http://terminology.hl7.org/CodeSystem/v3-NullFlavor",
          "code" : "UNK",
          "display" : "Unknown"
          }
      }
    }
    
    initial_ethnicity_json = {
      "extension" : [
        {
          "$ombCategory"
        },
        {
          "url" : "text",
          "valueString" : "$text"
        }
      ],
      "url" : "http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity"
    }
    #Create an ombcategory and detailed section of ethnicitiy extension
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value):
        """Assign an ethnicity extension value.

        Mirrors the behaviour of :class:`PatientRaceExtensionValueHandler` – if the
        supplied ``value`` matches one of the known ``omb_categories`` the
        appropriate extension is added/updated and ``final_struct`` is returned.
        When there is **no** match the function now returns an empty dictionary
        (``{}``) to satisfy the new test requirement.
        """
        original_final_struct = copy.deepcopy(final_struct)  # Keep a copy of the original structure
        # Retrieve the ethnicity extension if it exists; make it if it does not.
        if 'extension' not in final_struct:
            final_struct['extension'] = []
        ethnicity_block = utilFindExtensionWithURL(final_struct['extension'], 'http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity')
        if ethnicity_block is None:
            ethnicity_block = copy.deepcopy(self.initial_ethnicity_json)
            final_struct['extension'].append(ethnicity_block)

        matched = False
        for ethnicity_key, ethnicity_structure in self.omb_categories.items():
            if value.strip().lower() == ethnicity_key.strip().lower():
                matched = True
                # Replace $ombCategory in the extension list
                for i, item in enumerate(ethnicity_block["extension"]):
                    if isinstance(item, set) and "$ombCategory" in item:
                        # Replace the set with the new structure
                        ethnicity_block["extension"][i] = copy.deepcopy(ethnicity_structure)
                    elif isinstance(item, dict) and item.get("valueString") == "$text":
                        item['valueString'] = ethnicity_key
                return final_struct
        # No match – return empty dict as required
        if not matched:
            final_struct.clear()  # Clear any changes made to final_struct
            final_struct.update(original_final_struct)  # Revert to the original structure to discard any partial changes
            return {}
      
class PatientBirthSexExtensionValueHandler(AbstractStructureHandler):
    birth_sex_block = {
      "url" : "http://hl7.org/fhir/us/core/StructureDefinition/us-core-birthsex",
      "valueCode" : "$value"
    }
    #Assigna birthsex extension
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value):
        #Retrieve the birthsex extension if it exists; make it if it does not.
        if 'extension' not in final_struct:
            final_struct['extension'] = []
        birthsex_block = utilFindExtensionWithURL(final_struct['extension'], 'http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity')
        if birthsex_block is None:
            birthsex_block = self.birth_sex_block
            birthsex_block['valueCode'] = value
            final_struct['extension'].append(birthsex_block)
        pass
      
class PatientMRNIdentifierValueHandler(AbstractStructureHandler):
    patient_mrn_block = {
      "use" : "usual",
      "type" : {
        "coding" : [
          {
            "system" : "http://terminology.hl7.org/CodeSystem/v2-0203",
            "code" : "MR",
            "display" : "Medical Record Number"
          }
        ],
        "text" : "Medical Record Number"
      },
      "system" : "$system",
      "value" : "$value"
    }
    #Assign a MRN identifier
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value):
        #Retrieve the MRN identifier if it exists; make it if it does not.
        target_identifier = self.patient_mrn_block
        new_identifier = True
        if 'identifier' not in final_struct:
          final_struct['identifier'] = []
        for identifier in final_struct['identifier']:
          if 'type' in identifier:
            for coding in identifier['type']['coding']:
              if coding['code'] == 'MR':
                target_identifier = identifier
                new_identifier = False
        if new_identifier:
          final_struct['identifier'].append(target_identifier)
        target_identifier[key] = value
        pass
      
class PatientSSNIdentifierValueHandler(AbstractStructureHandler):
    patient_mrn_block = {
      "use" : "usual",
      "type" : {
        "coding" : [
          {
            "system" : "http://terminology.hl7.org/CodeSystem/v2-0203",
            "code" : "SS"
          }
        ],
        "text" : "Social Security Number"
      },
      "system" : "$system",
      "value" : "$value"
    }
    #Assign a MRN identifier
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value):
        #Retrieve the MRN identifier if it exists; make it if it does not.
        target_identifier = self.patient_mrn_block
        new_identifier = True
        if 'identifier' not in final_struct:
          final_struct['identifier'] = []
        for identifier in final_struct['identifier']:
          if 'type' in identifier:
            for coding in identifier['type']['coding']:
              if coding['code'] == 'SS':
                target_identifier = identifier
                new_identifier = False
        if new_identifier:
          final_struct['identifier'].append(target_identifier)
        target_identifier[key] = value
        pass
      
class OrganizationIdentiferNPIValueHandler(AbstractStructureHandler):
    npi_identifier_block = {
      "system" : "http://hl7.org.fhir/sid/us-npi",
      "value" : "$value"
    }
    #Assigna birthsex extension
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value):
        #Retrieve the birthsex extension if it exists; make it if it does not.
        if 'identifier' not in final_struct:
            final_struct['identifier'] = []
        identifier_block = next((entry for entry in final_struct['identifier'] if entry['system'] == "http://hl7.org.fhir/sid/us-npi"), None)
        if identifier_block is None:
          identifier_block = self.npi_identifier_block
          final_struct['identifier'].append(identifier_block)
        identifier_block['value'] = str(value)
        pass
      
class OrganizationIdentiferCLIAValueHandler(AbstractStructureHandler):
    clia_identifier_block = {
      "system" : "urn:oid:2.16.840.1.113883.4.7",
      "value" : "$value"
    }
    #Assign a birthsex extension
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value):
        #Retrieve the birthsex extension if it exists; make it if it does not.
        if 'identifier' not in final_struct:
            final_struct['identifier'] = []
        identifier_block = next((entry for entry in final_struct['identifier'] if entry['system'] == "urn:oid:2.16.840.1.113883.4.7"), None)
        if identifier_block is None:
          identifier_block = self.clia_identifier_block
          final_struct['identifier'].append(identifier_block)
        identifier_block['value'] = str(value)
        pass
      
class PractitionerIdentiferNPIValueHandler(AbstractStructureHandler):
    npi_identifier_block = {
      "system" : "http://hl7.org.fhir/sid/us-npi",
      "value" : "$value"
    }
    #Assigna birthsex extension
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value):
        #Retrieve the birthsex extension if it exists; make it if it does not.
        if 'identifier' not in final_struct:
            final_struct['identifier'] = []
        identifier_block = next((entry for entry in final_struct['identifier'] if entry['system'] == "http://hl7.org.fhir/sid/us-npi"), None)
        if identifier_block is None:
          identifier_block = self.npi_identifier_block
          final_struct['identifier'].append(identifier_block)
        identifier_block['value'] = value
        pass
      
class ObservationComponentHandler(AbstractStructureHandler):
    pulse_oximetry_oxygen_flow_rate = {
      "code" : {
        "coding" : [
          {
            "system" : "http://loinc.org",
            "code" : "3151-8",
            "display" : "Inhaled oxygen flow rate"
          }
        ],
        "text" : "Inhaled oxygen flow rate"
      }
    }
    pulse_oximetry_oxygen_concentration = {
      "code" : {
        "coding" : [
          {
            "system" : "http://loinc.org",
            "code" : "3150-0",
            "display" : "Inhaled oxygen concentration"
          }
        ],
        "text" : "Inhaled oxygen concentration"
      }
    }
    #Find the appropriate component for the observaiton; then call build_structure again to continue the drill down
    def assign_value(self, json_path, resource_definition, dataType, final_struct, key, value):
        #Check to make sure the component part exists
        if 'component' not in final_struct:
          final_struct['component'] = []
        components = final_struct['component']
        #Look through the qualifier parts.
        parts = json_path.split('.')
        key_part = parts[1][:parts[1].index('[')]
        qualifier = parts[1][parts[1].index('[')+1:parts[1].index(']')]
        qualifier_condition = qualifier.split('=')
        
        target_component: dict = {}
        if qualifier_condition[0] == 'code' and qualifier_condition[1] == '3151-8':
          target_component = findComponentWithCoding(components, '3151-8') or self.pulse_oximetry_oxygen_flow_rate
          if target_component is self.pulse_oximetry_oxygen_flow_rate:
            components.append(target_component)
        if qualifier_condition[0] == 'code' and qualifier_condition[1] == '3150-0':
          target_component = findComponentWithCoding(components, '3150-0') or self.pulse_oximetry_oxygen_concentration
          if target_component is self.pulse_oximetry_oxygen_concentration:
            components.append(target_component)
        #Recurse back down into 
        # current_struct: Dict, json_path: str, resource_definition: ResourceDefinition, dataType: str, parts: List[str], value: Any, previous_parts: List[str]
        return conversion.build_structure(target_component, '.'.join(parts[2:]), resource_definition, dataType, parts[2:], value, parts[:2])
        pass

#Special Handler just for $values. This one is data absent reason
class AbstractValueHandler(ABC):
  
  @abstractmethod
  def assign_value(self, final_struct, key, value, valueType):
    pass
  

class DataAbsentReasonHandler(AbstractValueHandler):
  #Assign data absent reason extension
    data_absent_reason_block = {
      "url" : "http://hl7.org/fhir/StructureDefinition/data-absent-reason",
      "value" : "$value"
    }
    data_absent_reason_values = ['$unknown','$asked-unknown','$temp-unknown','$not-asked','$asked-declined','$masked','$not-applicable','$unsupported','$as-text','$error','$not-a-number','$negative-infinity','$positive-infinity','$not-performed','$not-permitted']
    def assign_value(self, final_struct, key, value, valueType):
        #Trim the value so the '$' is missing
        if value and value.startswith('$'):
          value = value[1:]
        # if the final_struct is a list, we need the first entry of the list
        if isinstance(final_struct, list):
          if len(final_struct) == 0:
            temp = {}
            final_struct.append(temp)
            final_struct = temp
          else:
            final_struct = final_struct[0]
        #Retrieve the DAR extension if it exists; make it if it does not.
        if 'extension' not in final_struct:
            final_struct['extension'] = []
        data_absent_reason_block = utilFindExtensionWithURL(final_struct['extension'], 'http://hl7.org/fhir/StructureDefinition/data-absent-reason')
        if data_absent_reason_block is None:
            data_absent_reason_block = self.data_absent_reason_block
            data_absent_reason_block['value'] = value
            final_struct['extension'].append(data_absent_reason_block)
        pass
      
#Data dictionary of jsonpaths to match vs classes that need to be called
custom_structure_handlers = {
    "Patient.extension[Race].ombCategory": PatientRaceExtensionValueHandler(),
    "Patient.extension[Ethnicity].ombCategory": PatientEthnicityExtensionValueHandler(),
    "Patient.extension[Birthsex].value": PatientBirthSexExtensionValueHandler(),
    "Patient.identifier[type=MR].system": PatientMRNIdentifierValueHandler(),
    "Patient.identifier[type=MR].value": PatientMRNIdentifierValueHandler(),
    "Patient.identifier[type=MRN].system": PatientMRNIdentifierValueHandler(),
    "Patient.identifier[type=MRN].value": PatientMRNIdentifierValueHandler(),
    "Patient.identifier[type=SSN].system": PatientSSNIdentifierValueHandler(),
    "Patient.identifier[type=SSN].value": PatientSSNIdentifierValueHandler(),
    "Organization.identifier[system=NPI].value": OrganizationIdentiferNPIValueHandler(),
    "Organization.identifier[system=CLIA].value": OrganizationIdentiferCLIAValueHandler(),
    "Practitioner.identifier[system=NPI].value": PractitionerIdentiferNPIValueHandler(),
    "Observation.component[": ObservationComponentHandler()
}

#Data definition of values to match vs classes tht need to be called
custom_value_handlers = [
  {'value_criteria':DataAbsentReasonHandler.data_absent_reason_values, 'handler': DataAbsentReasonHandler()}
]


from typing import Any, Dict, List, Optional

from .common import get_value_from_keys


class ResourceDefinition:
    """
    A class to represent a Resource Definition for FHIR initialization.
    """
    entityName_keys: List[str] = ['Entity Name', 'name', 'entity_name']
    resourceType_keys: List[str] = ['ResourceType', 'resource_type', 'type']
    profile_keys: List[str] = ['Profile(s)', 'profiles', 'profile_list']
    
    def __init__(self, entityName: str, resourceType: str, profiles: List[str]):
        """
        Initializes the ResourceLink object from a dictionary.

        Args:
            data: A dictionary containing 'EntityName', 'ResourceType', and 'Profile(s)'.
        """
        self.entityName: str = entityName
        self.resourceType: str = resourceType
        self.profiles: List[str] = profiles
        
    @classmethod
    def from_dict(cls, data:  Dict[str, Any]):
        return cls(get_value_from_keys(data, cls.entityName_keys, ''), get_value_from_keys(data, cls.resourceType_keys, ''), get_value_from_keys(data, cls.profile_keys, []))

    def __repr__(self) -> str:
        return f"ResourceDefinition(entityName='{self.entityName}', resourceType='{self.resourceType}', profiles={self.profiles})"


from .model.cohort_data_entity import CohortData as CohortData
from .model.resource_definition_entity import ResourceDefinition as ResourceDefinition
from .model.resource_link_entity import ResourceLink as ResourceLink

def read_xlsx_and_process(file_path): ...
def process_sheet_resource_definitions(sheet) -> list[ResourceDefinition]: ...
def process_sheet_resource_links(sheet) -> list[ResourceLink]: ...
def process_sheet_patient_data_revised(sheet, resource_definition_entities): ...



import logging

class CustomLoggingFormatter(logging.Formatter):
    """Custom Formatter object for formatting logging messages throughout the API"""

    grey = "\x1b[38;21m"
    green = "\x1b[32m"
    yellow = "\x1b[33m"
    red = "\x1b[31m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format_str = "{asctime}   {levelname:8s} --- {name}: {message}"

    FORMATS = {
        logging.DEBUG: grey + format_str + reset,
        logging.INFO: green + format_str + reset,
        logging.WARNING: yellow + format_str + reset,
        logging.ERROR: red + format_str + reset,
        logging.CRITICAL: bold_red + format_str + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, "%m/%d/%Y %I:%M:%S %p", style="{")
        return formatter.format(record)
